#-----------------------------------------------------------------------------


#Optophone simulator
#
#by Ockham occam24@aol.com
#Feb 29, 2020

#-----------------------------------------------------------------------------
"""

Making a modern version of the Optophone reading device.  The Optophone

In digital form:

1. scan a horizontal band in the image.

2. Read pixels in a narrow vertical slot scanning across the band.

3. Mix the 5 tones into a WAV file according to the ink at 5 vertical
zones within the slot.

After this is working, see if it's possible to animate the model
along with the WAV.   Building the WAV is the hard part, so do first.

The mixing should be done in arrays first.  For each frequency, scan
the text and write the PercentBlack at each cell.
PercentBlack = ((256*3) - (r + g + b)) / (256*3)
Then use each of these 'envelope' arrays to modulate
a sine at the respective frequency; then add the
modulated sines together into final wave.


"""
#-----------------------------------------------------------------------------

from Tkinter import *
import tkFileDialog
import pickle
import os
import math
import wave
import array
from PIL import Image,ImageTk

#-----------------------------------------------------------------------------

global imFile,photoIm,resizedIm
global BMPname, WAVname
global fpWave
global DestArraySize
global DestArray
global FreqArraySize
global Freq1Array, Freq2Array,Freq3Array, Freq4Array,Freq5Array
global TopOfFirstY,TopOfSecondY  # for mouse marking
global FREQPART,HEIGHT5,HEIGHT4,HEIGHT3,HEIGHT2,HEIGHT1


BMPname = ''
WAVname = ''
fpWave = None
photoIm = None
resizedIm = None

DestArraySize = 0
DestArray = None



RadToDeg = 180.0 / math.pi
TwoPi = 2.0 * math.pi
RGBdenom = float(256 * 3)


#  For testing with the actual demo, I know the exact heights of the actual JPG.
#  Each line from ascender to descender is 142 pixels high, with 45 pixels between
#  them.  So the scanheight is 142 and the leadheight is 45.
# Later decided to skip the leads.

# Start these with defaults
TopOfFirstY = 0
TopOfSecondY = 170

WIDTH = 1024
HEIGHT = 1024
CELLHEIGHT = 170 # default unless set by mouse
CELLWIDTH = 1  # Constant.  At first I was trying to make this wider, but 1 is right.


# Arbitrarily making the total wave time 15 seconds regardless of pixels.

# Freqs used in original.  Freq 1 is lowest physically and musically.
# Freq1 basically reads descenders, Freq5 reads ascenders and UC.
FREQ1 = 334
FREQ2 = 512
FREQ3 = 576
FREQ4 = 640
FREQ5 = 763

# Heights for each cell, as ratios of main HEIGHT added from TOP of row:
# Pixels are numbered downward, so topmost cell is least added to top.
FREQPART = CELLHEIGHT / 5
HEIGHT5 = FREQPART * 0
HEIGHT4 = FREQPART * 1
HEIGHT3 = FREQPART * 2
HEIGHT2 = FREQPART * 3
HEIGHT1 = FREQPART * 4


Channels = 1
SampWidth = 2
SampRate = 44100

#-----------------------------------------------------------------------------

class App:
    def __init__(self, master, textMessage):
        self.master = master
        master.title("Optophone simulator")

        self.TopFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.TopFrame.grid(row=1,column=0)
        self.PicFrame=Frame(self.master,width=480,height=320,borderwidth=2,relief=RIDGE)
        self.PicFrame.grid(row=2,column=0)
        self.PicFrame.grid_propagate(0)
        self.BotFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.BotFrame.grid(row=3,column=0)

        self.Pic = Canvas(self.PicFrame,  width=512,height=512,bg="white",borderwidth=2,relief=RIDGE)
        self.Pic.grid(row=0,column=0)

        self.Pic.bind("<Button-1>",self.LpressAction)
        self.Pic.bind("<Button-3>",self.RpressAction)

        self.StatusEntry = Entry(self.TopFrame,width=70)
        self.StatusEntry.grid(row=0,column=0,columnspan=3)
        self.StatusEntry.insert(0,"First pick an image.")

        # Go/nogo buttons.
        self.ButtonName = Button(self.TopFrame, text="Pick image", command=self.HandleName)
        self.ButtonName.grid(row=1, column=0,sticky=E+W)

        self.ButtonWAV = Button(self.BotFrame, text="Make sound", command=self.HandleScan)
        self.ButtonWAV.grid(row=0, column=0)

        self.master.protocol('WM_DELETE_WINDOW', self.HandleCancel)

    # - - - - - - - - - - - - - - - - - -

    def LpressAction(self,event):  # Mark top of first line
        global TopOfFirstY
        TopOfFirstY = event.y

    # - - - - - - - - - - - - - - - - - -

    def RpressAction(self,event):  # Mark top of second line
        global TopOfFirstY,TopOfSecondY,CELLHEIGHT
        global FREQPART,HEIGHT5,HEIGHT4,HEIGHT3,HEIGHT2,HEIGHT1

        TopOfSecondY = event.y

        CELLHEIGHT = int( float(TopOfSecondY - TopOfFirstY) *  1024.0 / float(self.Pic.winfo_height()) )
        if CELLHEIGHT == 0: CELLHEIGHT=5 # min usable just in case
        FREQPART = CELLHEIGHT / 5
        HEIGHT5 = FREQPART * 0
        HEIGHT4 = FREQPART * 1
        HEIGHT3 = FREQPART * 2
        HEIGHT2 = FREQPART * 3
        HEIGHT1 = FREQPART * 4

        self.Pic.create_line(100,TopOfFirstY,200,TopOfSecondY,fill="red",width=3)

    # - - - - - - - - - - - - - - - - - -

    def ShowStatus(self,S): # Just saving repetition
        self.StatusEntry.delete(0,END)
        self.StatusEntry.insert(0,S)
        #self.StatusEntry.update_idletasks()

    # - - - - - - - - - - - - - - - - - -

    def HandleName(self):
        global imFile,photoIm,resizedIm
        global BMPname, WAVname
        global fpWave
        global TopOfFirstY,TopOfSecondY,CELLHEIGHT
        global DestArraySize
        global DestArray
        global FreqArraySize
        global Freq1Array, Freq2Array,Freq3Array, Freq4Array,Freq5Array

        # Bring in an image and store it via PIL.
        # Pop up file box to get name.
        BMPname=tkFileDialog.askopenfilename(defaultextension=[("BMP file", "*.BMP *.bmp")] )

        if not BMPname:
            self.ShowStatus("No file selected. Try again.")
            return 0

        # resize to fit canvas
        imFile = Image.open(BMPname)
        canvw=self.Pic.winfo_width()
        canvh=self.Pic.winfo_height()
        resizedIm = imFile.resize((canvw,canvh), Image.ANTIALIAS)
        photoIm = ImageTk.PhotoImage(resizedIm)
        imTag = self.Pic.create_image(0,0,anchor=NW,image=photoIm)


        # Now using the filename we create a sound output file and arrays.

        # Create a WAV file with same name.
        front,ext = os.path.splitext(BMPname)
        # Open for write
        BaseName = front
        fpWave = 0
        N = 0
        TryName = ""
        for N in range(0,100): # 100 should be enough; also avoids infinite loop.
            TryName = "%s%-.2d.WAV" % (BaseName,N)
            if not os.path.isfile(TryName):  # Doesn't exist. Open and stop looking.
                fpWave = wave.open(TryName,"wb")
                break

        WAVname = TryName
        # Implied else: Exists.  Keep running with increasing N.

        if not fpWave:
            raise Exception("Wave file %s failed to open." % WAVname)

        # Set up the header
        fpWave.setnchannels(Channels)
        fpWave.setsampwidth(SampWidth)
        fpWave.setframerate(SampRate)

        # For one line in our test image we arbitrarily use 15 seconds...
        # DestArraySize is 15 seconds of wave.  It's NOT the same as the
        # arrays of modulating signal.  Those arrays (FreqNarray) are
        # just large enough to hold the number of 'reflections' in the
        # pattern of scans.  For now, just two rows with 1024 points in each.
        #
        DestArraySize = int(SampRate * 15) # arbitrary
        DestArray       = array.array('h',[0]*DestArraySize)
        # Separate arrays for each freq, corresponding to horizontal scan of one line.
        # As we scan one line, we handle each 'floor' separately, counting black pixels
        # within the 'floor'.  When done with all 5 'floors', we mix them into the wave.
        # These are arrays of 'modulating signal', floating point 0 to 1.
        FreqArraySize = WIDTH * 3  # For now.  Should be WIDTH * (HEIGHT/(SCANHEIGHT+LEADHEIGHT))
        Freq1Array = array.array('f',[0.0]*FreqArraySize)
        Freq2Array = array.array('f',[0.0]*FreqArraySize)
        Freq3Array = array.array('f',[0.0]*FreqArraySize)
        Freq4Array = array.array('f',[0.0]*FreqArraySize)
        Freq5Array = array.array('f',[0.0]*FreqArraySize)


        self.ShowStatus("Leftclick at top of 1st line, rightclick at top of 2nd line, then Make Sound.")

    # - - - - - - - - - - - - - - - - - -

    def HandleScan(self):
        global imFile
        global BMPname, WAVname
        global fpWave
        global TopOfFirstY,TopOfSecondY,CELLHEIGHT
        global FREQPART,HEIGHT5,HEIGHT4,HEIGHT3,HEIGHT2,HEIGHT1
        global DestArraySize
        global DestArray
        global FreqArraySize
        global Freq1Array, Freq2Array,Freq3Array, Freq4Array,Freq5Array

        self.ShowStatus("scanning picture")
        arrayindex = -1
        # Limit to first tHREE rows for now
        # Now that the pic is showing, let's try to trace the action on the pic.
        # Good for animation, and also good to tell what's happening.
        for startY in range( 0, (3*CELLHEIGHT), CELLHEIGHT ): # stepping down lines.  Three lines arbitrarily.
            for startX in range( 0, WIDTH): # stepping across one line
                # Show this vertical scan on the pic:
                self.Pic.create_line((startX/2),(startY/2),(startX/2),((startY+CELLHEIGHT)/2),fill="green",width=3)
                self.Pic.update_idletasks()

                arrayindex = arrayindex+1

                # Run each freq separately, not looped.
                # First Freq5, top 'floor':
                rsum,gsum,bsum=(0,0,0)  # rgb tuple
                for cellY in range((startY+HEIGHT5),(startY+HEIGHT5+FREQPART)):
                    # get one pixel
                    xy=(startX,cellY)
                    (r,g,b) = imFile.getpixel(xy)
                    rsum = rsum + r
                    gsum = gsum + g
                    bsum = bsum + b
                # Get mean of this cell and place in Freq5Array:
                meanR = float(rsum) / float(FREQPART)
                meanG = float(gsum) / float(FREQPART)
                meanB = float(bsum) / float(FREQPART)
                PercentBlack = (RGBdenom - (meanR + meanG + meanB)) / RGBdenom
                Freq5Array[arrayindex] = PercentBlack

                # Now Freq4, next to top 'floor':
                rsum,gsum,bsum=(0,0,0)  # rgb tuple
                for cellY in range((startY+HEIGHT4),(startY+HEIGHT4+FREQPART)):
                    # get one pixel
                    xy=(startX,cellY)
                    rgb = imFile.getpixel(xy)
                    rsum = rsum + r
                    gsum = gsum + g
                    bsum = bsum + b
                # Get mean of this cell and place in Freq5Array:
                meanR = float(rsum) / float(FREQPART)
                meanG = float(gsum) / float(FREQPART)
                meanB = float(bsum) / float(FREQPART)
                PercentBlack = (RGBdenom - (meanR + meanG + meanB)) / RGBdenom
                Freq4Array[arrayindex] = PercentBlack

                # nOW Freq3 in the middle:
                rsum,gsum,bsum=(0,0,0)  # rgb tuple
                for cellY in range((startY+HEIGHT3),(startY+HEIGHT3+FREQPART)):
                    # get one pixel
                    xy=(startX,cellY)
                    rgb = imFile.getpixel(xy)
                    rsum = rsum + r
                    gsum = gsum + g
                    bsum = bsum + b
                # Get mean of this cell and place in Freq5Array:
                meanR = float(rsum) / float(FREQPART)
                meanG = float(gsum) / float(FREQPART)
                meanB = float(bsum) / float(FREQPART)
                PercentBlack = (RGBdenom - (meanR + meanG + meanB)) / RGBdenom
                #print PercentBlack #degug
                Freq3Array[arrayindex] = PercentBlack

                # nOW Freq2 next to bottom:
                rsum,gsum,bsum=(0,0,0)  # rgb tuple
                for cellY in range((startY+HEIGHT2),(startY+HEIGHT2+FREQPART)):
                    # get one pixel
                    xy=(startX,cellY)
                    rgb = imFile.getpixel(xy)
                    rsum = rsum + r
                    gsum = gsum + g
                    bsum = bsum + b
                # Get mean of this cell and place in Freq5Array:
                meanR = float(rsum) / float(FREQPART)
                meanG = float(gsum) / float(FREQPART)
                meanB = float(bsum) / float(FREQPART)
                PercentBlack = (RGBdenom - (meanR + meanG + meanB)) / RGBdenom
                Freq2Array[arrayindex] = PercentBlack

                # Finally Freq1 at bottom:
                rsum,gsum,bsum=(0,0,0)  # rgb tuple
                for cellY in range((startY+HEIGHT1),(startY+HEIGHT1+FREQPART)):
                    # get one pixel
                    xy=(startX,cellY)
                    rgb = imFile.getpixel(xy)
                    rsum = rsum + r
                    gsum = gsum + g
                    bsum = bsum + b
                # Get mean of this cell and place in Freq5Array:
                meanR = float(rsum) / float(FREQPART)
                meanG = float(gsum) / float(FREQPART)
                meanB = float(bsum) / float(FREQPART)
                PercentBlack = (RGBdenom - (meanR + meanG + meanB)) / RGBdenom
                Freq1Array[arrayindex] = PercentBlack

            # end loops on startX = one row of type.
        # end outer loop on startY for all rows

        # Now we have the 5 freq arrays for one row of type.


        self.BuildWave()  # separated out for readability


    # - - - - - - - - - - - - - - - - - -

    def BuildWave(self):
        global imFile
        global BMPname, WAVname
        global fpWave
        global DestArraySize
        global DestArray
        global Freq1Array, Freq2Array,Freq3Array, Freq4Array,Freq5Array

        # Note that we don't need a separate sin-holder array
        self.ShowStatus("working on wave")

        # This multiplier cuts each modulator value by 1/5 so the
        # sum will stay inside 1, then multiplies by the DAC value.
        LevelMult = 32676.0 / 5.0

        # Since we're expanding the relatively short FreqArrays into
        # the much longer WAV, we need to use one value from FreqArray
        # for a long interval in the WAV.  The interval is DestArraySize/FreqArraySize.
        Span = int(DestArraySize / FreqArraySize)


        # First Freq5
        SamplesPerCycle = SampRate / float(FREQ5)
        # Figure the angle increment for each sampling point:
        #    Radians / Sample = radians/Cycle * Cycles / Sample.
        # or Radians / Sample = TwoPi         / (SamplesPerCycle)
        RadStep = TwoPi / float(SamplesPerCycle)

        # Form a sin wave at FREQ5.
        # Modulate this by the Freq5Array and add it into the big wave.
        for i in range(DestArraySize):
            # Get radians corresponding to this index:
            Rads = float(i) * RadStep
            # Get corresponding level from Freq5Array
            # Put sin value in the buffer (at int index)

            # Reuse the 'spanned' index in FreqArray
            FreqIndex = int(i / Span)
            if FreqIndex > (FreqArraySize-2): break

            SinVal = int(LevelMult * Freq5Array[FreqIndex] * math.sin(Rads))
            # Add this to the big destarray
            DestArray[i] = DestArray[i] + SinVal
        # end adding Freq5 values



        # Then Freq4
        SamplesPerCycle = SampRate / float(FREQ4)
        # Figure the angle increment for each sampling point:
        #    Radians / Sample = radians/Cycle * Cycles / Sample.
        # or Radians / Sample = TwoPi         / (SamplesPerCycle)
        RadStep = TwoPi / float(SamplesPerCycle)

        # Form a sin wave at FREQ5.
        # Modulate this by the Freq5Array and add it into the big wave.
        for i in range(DestArraySize):
            # Get radians corresponding to this index:
            Rads = float(i) * RadStep
            # Get corresponding level from Freq5Array
            # Put sin value in the buffer (at int index)

            # Reuse the 'spanned' index in FreqArray
            FreqIndex = int(i / Span)
            if FreqIndex > (FreqArraySize-2): break

            SinVal = int(LevelMult * Freq4Array[FreqIndex] * math.sin(Rads))
            # Add this to the big destarray
            DestArray[i] = DestArray[i] + SinVal
        # end adding Freq5 values


        # Then Freq3
        SamplesPerCycle = SampRate / float(FREQ3)
        # Figure the angle increment for each sampling point:
        #    Radians / Sample = radians/Cycle * Cycles / Sample.
        # or Radians / Sample = TwoPi         / (SamplesPerCycle)
        RadStep = TwoPi / float(SamplesPerCycle)

        # Form a sin wave at FREQ5.
        # Modulate this by the Freq5Array and add it into the big wave.
        for i in range(DestArraySize):
            # Get radians corresponding to this index:
            Rads = float(i) * RadStep
            # Get corresponding level from Freq5Array
            # Put sin value in the buffer (at int index)

            # Reuse the 'spanned' index in FreqArray
            FreqIndex = int(i / Span)
            if FreqIndex > (FreqArraySize-2): break

            SinVal = int(LevelMult * Freq3Array[FreqIndex] * math.sin(Rads))
            # Add this to the big destarray
            DestArray[i] = DestArray[i] + SinVal
        # end adding Freq5 values


        # Then Freq2
        SamplesPerCycle = SampRate / float(FREQ2)
        # Figure the angle increment for each sampling point:
        #    Radians / Sample = radians/Cycle * Cycles / Sample.
        # or Radians / Sample = TwoPi         / (SamplesPerCycle)
        RadStep = TwoPi / float(SamplesPerCycle)

        # Form a sin wave at FREQ5.
        # Modulate this by the Freq5Array and add it into the big wave.
        for i in range(DestArraySize):
            # Get radians corresponding to this index:
            Rads = float(i) * RadStep
            # Get corresponding level from Freq5Array
            # Put sin value in the buffer (at int index)

            # Reuse the 'spanned' index in FreqArray
            FreqIndex = int(i / Span)
            if FreqIndex > (FreqArraySize-2): break

            SinVal = int(LevelMult * Freq2Array[FreqIndex] * math.sin(Rads))
            # Add this to the big destarray
            DestArray[i] = DestArray[i] + SinVal
        # end adding Freq5 values


        # Finally Freq1
        SamplesPerCycle = SampRate / float(FREQ1)
        # Figure the angle increment for each sampling point:
        #    Radians / Sample = radians/Cycle * Cycles / Sample.
        # or Radians / Sample = TwoPi         / (SamplesPerCycle)
        RadStep = TwoPi / float(SamplesPerCycle)

        # Form a sin wave at FREQ5.
        # Modulate this by the Freq5Array and add it into the big wave.
        for i in range(DestArraySize):
            # Get radians corresponding to this index:
            Rads = float(i) * RadStep
            # Get corresponding level from Freq5Array
            # Put sin value in the buffer (at int index)

            # Reuse the 'spanned' index in FreqArray
            FreqIndex = int(i / Span)
            if FreqIndex > (FreqArraySize-2): break

            SinVal = int(LevelMult * Freq1Array[FreqIndex] * math.sin(Rads))
            # Add this to the big destarray
            DestArray[i] = DestArray[i] + SinVal
        # end adding Freq5 values



        # Write DestArray to fpWave
        fpWave.writeframes(DestArray.tostring())
        fpWave.close()

        self.ShowStatus("done")

    # - - - - - - - - - - - - - - - - - -


    # - - - - - - - - - - - - - - - - - -

    def HandleCancel(self):
        self.master.destroy()

#------------------------------------------------------------
# Activate the loop
#------------------------------------------------------------

root = Tk()
app = App(root, '')
root.mainloop()


#------------------------------------------------------------
# End TK loop.
#-----------------------------------------------------------------------------

